# Generated by Django 2.2.27 on 2022-06-20 10:37

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("audoma_api", "0003_merge_20220526_1439"),
        ("audoma_api", "0003_car_manufacturer"),
    ]

    operations = []
